@@ -42,6 +42,7 @@ static gchar *contributors[] = {
 	"Marcus Brubaker",
 	"Christian Bucher",
 	"Ross Burton",
+	"Brian Cameron",
 	"Dave Camp",
 	"Didier Carlier",
 	"Anders Carlsson",
@@ -60,6 +61,7 @@ static gchar *contributors[] = {
 	"Andrew Clausen",
 	"Jeremy Collins",
 	"Rusty Conover",
+	"Pat Costello",
 	"Alan Cox",
 	"Dennis Cranston",
 	"Mark Crichton",
@@ -259,6 +261,7 @@ static gchar *contributors[] = {
 	"Lars Rydlinge",
 	"Peter Ryland",
 	"Bibek Sahu",
+	"Arvind Samptur",
 	"Pablo Saratxaga",
 	/* Name in Japanese is: U7b39 U5c71 U548c U5b8f */
 	N_("Kazuhiro Sasayama"),
@@ -320,6 +323,7 @@ static gchar *contributors[] = {
 	"Bruno Widmann",
 	"Robert Wilhelm",
 	"Sebastian Wilhelmi",
+	"James Willcox",
 	"David Winkler",
 	"Jeremy Wise",
 	"Roger Wolff",
